#!/usr/bin/perl -w
# $Header: /home/name/projects/tv/RCS/tv-dl.pl,v 1.44 2012/12/16 21:29:40 name Exp $
#
# tv-dl.pl
#
#	TV program downloader
#
#	(c)2005 Hiro Sugawara
#	Distributed under terms of GPL 2.0
#

use strict;
use POSIX;
use File::Basename;
use Sys::Syslog;

my $myname	= basename $0;
my $cwd		= `pwd`;
chomp $cwd;
my $mydir	= dirname $0;
if ($mydir eq '.') {
	$mydir = '';
}
unless ($mydir =~ /^\//) {
	$mydir	= "$cwd/$mydir";
}
$mydir	=~ s/\/$//;
my $cfile	= "$mydir/conf/tv-dl.conf";

openlog("$myname\[$$\]", "", "local0");

sub help {
	print <<_END;
Usage: $myname [-c config][-d]
    -c config      configuration file ($cfile)
    -d             daemon mode
_END
}

my $speedlog	= "/var/log/6ga/speed.log";

# Forward declaration of config parameters
my $VIDEOSAVEDIR;
my $POLLINTERVAL;
my $ROOTURL;
my $QUERYURL;
my $STREAMURL;	# Unused
my $CHANNELURL;	# Unused
my $SSHKEYFILE;
my $SSHPORT;

# Hard-wired location for config file
my $daemon	= 0;

#---------------------------------------------------------------
# Parse options
#---------------------------------------------------------------
while ($#ARGV >= 0 && $ARGV[0] =~ /^-/) {
	my $opt	= shift @ARGV;
	if ($opt eq '-c') {
		$cfile	= shift @ARGV;
	} elsif ($opt eq '-d') {
		$daemon	= 1;
	} elsif ($opt eq '-h' || $opt eq '-?') {
		help();
		exit 0;
	} else {
		die "Unknown option $opt\n";
	}
}

readConfigFile($cfile);

#-------------------------------------------------------------
# Check options
#-------------------------------------------------------------
defined($VIDEOSAVEDIR)	|| die "VIDEOSAVEDIR undefined";
chdir $mydir		|| die "can't chdir to $mydir";
if (defined($POLLINTERVAL)) {
	($POLLINTERVAL =~ /^\d+$/) || die "bad POLLINTERVAL";
}
(defined($QUERYURL) && $QUERYURL =~ /^http/)	|| die "bad QUERYURL";
(defined($SSHPORT) && $SSHPORT =~ /^\d+$/)		|| die "bad SSHPORT";
defined($SSHKEYFILE) 	|| die "SSHKEYFILE undefined";
unless ($SSHKEYFILE =~ /^\//) {
	$SSHKEYFILE	= "$mydir/$SSHKEYFILE";
}

chdir $VIDEOSAVEDIR	|| die "can't chdir to $VIDEOSAVEDIR";
(-d 'comments')	|| mkdir 'comments'	|| die "can't mkdir comments";
(-d '.tmp')		|| mkdir '.tmp'	|| die "can't mkdir .tmp";
chdir '.tmp'	|| die "can't chdir to .tmp";

(-r $SSHKEYFILE) 	|| die "bad SSHKEYFILE $SSHKEYFILE";

my $userpasswd;
if ($QUERYURL =~ /^(http|https):\/\/([^:]+:[^\@]+)\@/) {
	$userpasswd	= $2;
}

#-------------------------------------------------------------
# Only a single process can run
#-------------------------------------------------------------
if ($daemon) {
	(fork > 0) && exit 0;
	close STDOUT;
	close STDERR;
	open  STDOUT, ">/dev/null";
	open  STDERR, ">&STDOUT";

	#
	# Reopen log with a new PID (child)
	#
	closelog();
	openlog("$myname\[$$\]", "", "local0");
}


my $lockfile	= $ENV{HOME} . "/lock/$myname";
$lockfile	=~ s/\.[^.]*$/.lock/;

umask 0377;
until (open LOCKFILE, ">$lockfile") {
	open(LOCKFILE, $lockfile) || die "can't open lockfile $lockfile for read\n";
	my $pid = <LOCKFILE>;
	close LOCKFILE;
	chomp $pid;
	if (-d "/proc/$pid") {
		exit 0;
	}
	unlink $lockfile	|| die "can't delete stale lockfile $lockfile";
}
print LOCKFILE "$$\n";
close LOCKFILE;

logit("Starting");
my $browser	= "elinks -no-home -source";

#============================================================
# Return iEPG contents in passed hash reference
#============================================================
sub readRemiEPG {
	my ($iepgref, $tvfile)	= @_;
	my $url	= $QUERYURL;

	if (defined($tvfile)) {
		$url	.= "?sched=$tvfile";
	}
	open(ELINKS, "$browser $url 2>/dev/null |") || return 0;

	while (<ELINKS>) {
		chomp;
		next if (/^\s*$/);
		my ($key, $value)	= split /:\s+/, $_, 2;
		$$iepgref{$key}	= $value;
	}
	close ELINKS;
	return 1;
}

#-------------------------------------------------------------
# Loop for all downloadable programs
#-------------------------------------------------------------
while (1) {
	my %iepg;

	unless (readRemiEPG(\%iepg)) {
		sleep $POLLINTERVAL;
		next;
	}

	my $xfrdurl	= $iepg{'tv-xfrd-url'};
	if (defined($userpasswd) && defined($xfrdurl)
		&& $xfrdurl !~ /^(http|https):\/\/([^:]+:[^\@]+)\@/) {
		$xfrdurl =~ s/^(http|https):\/\//$1:\/\/$userpasswd\@/;
	}
	my $dlurl	= $iepg{'tv-dl-url'};
	my $tvuser	= $iepg{'tv-user'};
	my $tvfile	= $iepg{'tv-file-name-base'};
	my $tvfsize	= $iepg{'file-size'};

	unless (defined($dlurl)) {
		last unless defined($POLLINTERVAL);
		sleep $POLLINTERVAL;
		next;
	}

	# Clean up broken files 3 days old or more
	delete_stalefiles();

	my $fbname	= sprintf "%04d%02d%02d%02d%02d",
							$iepg{year}, $iepg{month}, $iepg{date},
							split(/:/, $iepg{start});
	#---------------------------------------------------------
	# Declare starting of transfer
	#---------------------------------------------------------
	my $xferstart	= time;
	my $moreretry = 20;
	my $ksize  = 0;

	while ($moreretry-- > 0) {
		#---------------------------------------------------------
		# Set up handler for child exit
		#---------------------------------------------------------
		my $childexited	= 0;
		my $pid;

		logit("starting to download $tvfile ($dlurl)");
		##logit("rsync -az --partial-dir=.rsync-partial --timeout=600 --rsh=\"ssh -p $SSHPORT -l $tvuser -i $SSHKEYFILE\" $dlurl $tvfile.avi");
		if (!($pid = fork)) {
			if (!exec "rsync -az --partial-dir=.rsync-partial --timeout=600 --rsh=\"ssh -C -p $SSHPORT -l $tvuser -i $SSHKEYFILE\" $dlurl $tvfile.avi >/dev/null 2>&1") {
				logdie("failed to exec rsync: $!");
			}
		} elsif ($pid < 0) {
			logdie("failed to fork child: $!");
		}

		umask 0133;
		open(SPEEDLOG, ">>$speedlog");

		my $MOVINGAVG	= 4;
		my @pksize	= (0);
		my @ptime	= (time);

		if (-e "$tvfile.avi") {
			defined($ksize = fsizeInKB("$tvfile.avi"))
				|| logdie("fsizeInKB() for $tvfile.avi failed");
			$pksize[0]	= $ksize;
		}
		system "$browser \"$xfrdurl$ksize\" >/dev/null 2>&1";
		while (waitpid($pid, WNOHANG) <= 0) {
			sleep 17;
			my $ctime	= time;
			defined($ksize = fsizeInKB(".$tvfile.avi.*"))	|| next;
			push @pksize, $ksize;
			push @ptime, $ctime;
			while ($#pksize >= $MOVINGAVG) {
				shift @pksize;
			}
			while ($#ptime >= $MOVINGAVG) {
				shift @ptime;
			}
			if ($ptime[0] < $ctime && $pksize[0] < $ksize) {
				my $rate = sprintf '%.2f',
								($ksize - $pksize[0])/($ctime - $ptime[0]);
				system "$browser '$xfrdurl$ksize\@$rate' >/dev/null 2>&1";
				printf SPEEDLOG "%.24s %s\n", ctime(time), $rate;
			} else {
				system "$browser '$xfrdurl$ksize' >/dev/null 2>&1";
			}
		}
		close SPEEDLOG;

		#---------------------------------------------------------
		# Child has exited. Check status.
		#---------------------------------------------------------
		my $logmsg;
		my $fsize = fsizeInB("$tvfile.avi");
		if ($? & 127) {
			my $signo	= $? & 127;
			if ($signo == 15) {
				logit("%s: xfer stopped - aborting", $tvfile);
				exit 0;
			}
			$logmsg	= sprintf "%s: rsync died with signal %d, with%s coredump",
							$tvfile, $signo, ($? & 128) ? '' : 'out';
		} elsif (($? >> 8) == 23) {
			$logmsg	= "rsync succeeded in partial transfer";
		} elsif (($? >> 8) != 0) {
			$logmsg	= sprintf "rsync exited with value %d", $? >> 8;
		} elsif ($fsize ne $tvfsize) {
			$logmsg	= sprintf "file size %s < %s", $fsize, $tvfsize;
		} else {
			last;
		}

		if (!$moreretry) {
			logdie("$logmsg for $tvfile; too many retries - aborting");
		}

		unless (readRemiEPG(\%iepg, $tvfile)) {
			logdie("no iEPG for $tvfile anymore - aborting");
		}
		if ($iepg{'tv-stat'} ne 'Ready') {
			logdie("transfer of $tvfile aborted");
		}

		sleep 10;
		logit("$logmsg for $tvfile; retrying (%d more retries)", $moreretry);
	}

	#---------------------------------------------------------
	# Declare completion of transfer
	#---------------------------------------------------------
	system "$browser \"$xfrdurl-1\" >/dev/null 2>&1";
	$ksize = fsizeInKB("$tvfile.avi");
	my $elapsed	= time - $xferstart;
	logit("$tvfile: %dMB transferred in %dm%ss (%dkB/s)",
				$ksize/1000, $elapsed/60, $elapsed%60, $ksize/$elapsed);

	#---------------------------------------------------------
	# Rename to unique yyyymmddhhmmx.avi
	#---------------------------------------------------------
	my $inx;
	for ($inx = 1; $inx <= 9; $inx++) {
		last unless (-f "../$fbname$inx.avi");
	}

	if ($inx > 9) {
		### FIXME ###
	}

	$fbname	.= "$inx";
	rename "$tvfile.avi", "../$fbname.avi";
	chmod 0644, "../$fbname.avi";

	umask 0113;
	if (open CFILE, ">../comments/$fbname.txt") {
		printf CFILE "%s %s (%s)\n",
			defined($iepg{'program-title'}) ? $iepg{'program-title'} : '???',
			defined($iepg{'program-subtitle'})
					? $iepg{'program-subtitle'}
					: defined($iepg{'performer'})
						? $iepg{'performer'} : '',
			$iepg{station};
		close CFILE;
	}

	if (open IFILE, ">../comments/$fbname.iepg") {
		foreach my $key (keys(%iepg)) {
			printf IFILE "%s: %s\n", $key, $iepg{$key};
		}
		close IFILE;
	}
}

0;

#============================================================
# Delete stale broken files
#============================================================
sub delete_stalefiles{
	if (opendir DIR, '.') {
		while (my $file = readdir DIR) {
			if (-f $file
				&& $file =~ /\.avi/
				&& time - (stat($_))[9] > 7 * 24 * 60 * 60) {
				unlink $file;
				logit("stale broken file $file in .tmp removed");
			}
		}
		closedir DIR;
	}
}

#============================================================
# Log and die
#============================================================
sub logdie {
	my ($fmt, @args)	= @_;
	syslog('info', "$fmt", @args);
	die sprintf($fmt, @args);
}

#============================================================
# Log information
#============================================================
sub logit {
	my ($fmt, @args)	= @_;
	syslog('info', "$fmt", @args);
}

#============================================================
# Read in cofiguration file
#============================================================
sub readConfigFile {
	my ($configfile)	= @_;
	(defined $configfile)		|| die "no config file\n";

	open(CONFIG, $configfile)	|| die "config file $configfile not found/\n";
	while (<CONFIG>) {
		chomp;
		eval;
	}
	close CONFIG;
}

#============================================================
# Find file size in bytes
#============================================================
sub fsizeInB {
	my ($file)	= @_;
	my $size;

	if (open LS, "ls -l $file 2>/dev/null |") {
		$size	= <LS>;
		close LS;
		$size	= (split / /, $size)[4];
	}
	return $size;
}

#============================================================
# Find file size in kilo bytes
#============================================================
sub fsizeInKB {
	my ($file)	= @_;
	my $size;

	if (open LS, "ls -l --block-size=1000 $file 2>/dev/null |") {
		$size	= <LS>;
		close LS;
		$size	= (split / /, $size)[4];
	}
	return $size;
}
